mod bootstrap;
mod lsblk;
mod lvm;

use anyhow::format_err;
use log::error;
use std::io;
use std::process::{exit, Command};
use std::{env, fs};

fn main() {
    if let None = env::var_os("RUST_LOG") {
        env::set_var("RUST_LOG", "info");
    }
    env_logger::init();

    env::set_var("PATH", "/bin:/sbin");

    if let Some(arg) = env::args().into_iter().nth(1) {
        match arg.as_str() {
            "hello" => println!("hello"),
            _ => {
                error!("invalid arg: {arg}");
                exit(1);
            }
        };
        exit(0);
    }

    let cfg: bootstrap::config::Config = retry(|| {
        let cfg = fs::read("config.yaml").map_err(|e| format_err!("failed to read config: {e}"))?;
        serde_yaml::from_slice(cfg.as_slice())
            .map_err(|e| format_err!("failed to parse config: {e}"))
    });

    serde_yaml::to_writer(io::stdout(), &cfg).unwrap();

    mount("none", "/proc", "proc", None);
    mount("none", "/sys", "sysfs", None);
    mount("none", "/dev", "devtmpfs", None);
    mount("none", "/dev/pts", "devpts", Some("gid=5,mode=620"));

    exec("ash", vec!["-c", "set -ex; pvs; false"]);

    for pv in lvm::pvs().unwrap() {
        println!("pv: {pv:?}");
    }
    for vg in lvm::vgs().unwrap() {
        println!("vg: {vg:?}");
    }
    for lv in lvm::lvs().unwrap() {
        println!("lv: {lv:?}");
    }

    println!("lsblk: {:?}", lsblk::report());

    // TODO
}

fn mount(src: &str, dst: &str, fstype: &str, opts: Option<&str>) {
    let _ = fs::create_dir_all(dst).inspect_err(|e| error!("failed to create dir {dst}: {e}"));

    let mut args = vec![src, dst, "-t", fstype];
    if let Some(opts) = opts {
        args.extend(["-o", opts]);
    }
    exec("mount", args);
}

fn exec(prog: &str, args: Vec<&str>) {
    let cmd_str = sh_str(prog, &args);
    retry_or_ignore(|| {
        println!("\x1b[34m# {cmd_str}\x1b[0m");
        let mut cmd = Command::new(prog);
        cmd.args(args.clone());
        let s = cmd.status()?;
        if s.success() {
            Ok(())
        } else {
            Err(format_err!("command failed: {s}"))
        }
    });
}

fn retry_or_ignore(action: impl Fn() -> anyhow::Result<()>) {
    loop {
        match action() {
            Ok(_) => return,
            Err(e) => {
                eprintln!("\x1b[31m{e}\x1b[0m");

                loop {
                    eprint!("[r]etry, [i]gnore, or [s]hell? ");

                    let mut line = String::new();
                    io::stdin().read_line(&mut line).unwrap();

                    match line.trim() {
                        "r" => break,
                        "i" => return,
                        "s" => {
                            exec_shell();
                            break;
                        }
                        v => {
                            eprintln!("invalid choice: {v:?}");
                        }
                    };
                }
            }
        }
    }
}

fn retry<T>(action: impl Fn() -> anyhow::Result<T>) -> T {
    loop {
        match action() {
            Ok(v) => return v,
            Err(e) => {
                eprintln!("\x1b[31m{e}\x1b[0m");

                loop {
                    eprint!("[r]etry, or [s]hell? ");

                    let mut line = String::new();
                    io::stdin().read_line(&mut line).unwrap();

                    match line.trim() {
                        "r" => break,
                        "s" => {
                            exec_shell();
                            break;
                        }
                        v => {
                            eprintln!("invalid choice: {v:?}");
                        }
                    };
                }
            }
        }
    }
}

fn exec_shell() {
    let mut child = match Command::new("ash").spawn() {
        Ok(c) => c,
        Err(e) => {
            error!("failed to exec shell: {e}");
            return;
        }
    };

    let _ = child.wait();
}

fn sh_str(prog: &str, args: &Vec<&str>) -> String {
    use std::borrow::Cow;

    let mut buf = String::new();

    buf.push_str(&shell_escape::escape(Cow::from(prog)));

    for &arg in args {
        buf.push(' ');
        buf.push_str(&shell_escape::escape(Cow::from(arg)));
    }

    buf
}
